-- Filename:     seq_mult_2_fsmd.vhd
-- Filetype:     VHDL code (FSM with data)
-- Date:         31 may 2013
-- Update:       15 may 2014
-- Description:  FSMD description of the multiplier
-- Author:       J. op den Brouw
-- State:        demo
-- Error:        -
-- Version:      1.0.1
-- Copyright:    (c)2014, De Haagse Hogeschool

-- This VHDL description shows how to construct a
-- sequential multiplier using the known shift-and-add
-- method. This version uses a state machine with data
-- processing (a.k.a FSMD).
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- The Sequential Mulitplier Version 2 as FSM with data.
-- The operation is: resultp = dataa * datab
entity seq_mult_2_fsmd is
	generic (n : integer := 4);
	port (clk : in std_logic;
			areset : in std_logic;
			start : in std_logic;
			dataa : in unsigned(n-1 downto 0);
			datab : in unsigned(n-1 downto 0);
			resultp : out unsigned(2*n-1 downto 0);
			ready : out std_logic
	);
end seq_mult_2_fsmd;

-- This architecture uses variables for internal storage
architecture rtl_var of seq_mult_2_fsmd is
type state_type is (rest, iterate, done);
signal state : state_type;
begin

	-- The next state logic and register in one process
	-- together with the data processing
	process (clk, areset) is
	variable rega : unsigned (2*n-1 downto 0);
	variable regb : unsigned (n-1 downto 0);
	variable regp : unsigned (2*n-1 downto 0);
	constant n_zeros : unsigned (n-1 downto 0) := (others => '0');
	begin
		-- Asynchronous reset, set reset state, clear all registers
		if areset = '1' then
			state <= rest;
			rega := (others => '0');
			regb := (others => '0');
			regp := (others => '0');
      -- On positive edge, hence flipflops
		elsif rising_edge(clk) then
			case state is
				-- Wait for start to be asserted, the goto iterate
				-- Meanwhile, clear all registers
				when rest => if start = '1' then state <= iterate; end if;
								 rega := n_zeros & dataa;
								 regb := datab;
								 regp := (others => '0');
            -- As long as B is not 0, stay iterating, else we're done
            -- Increment P if lsb of B is 1, left shift A, right shift B
				when iterate => if regb = n_zeros then state <= done; end if;
								    if regb(0) = '1' then regp := regp + rega; end if;
									 rega := rega(2*n-2 downto 0) & '0';
				                regb := '0' & regb(n-1 downto 1);
				-- Wait for start to be de-asserted, then goto rest
				when done => if start = '0' then state <= rest; end if;

				-- If something goes horrible wrong, do nothing...
				when others => null;
			end case;
		end if;
		-- Output the result
		resultp <= regp;
	end process;

	-- Signal ready (combinatorial process)
	ready <= '1' when state = done else '0';

end architecture rtl_var;

-- This architecture uses signals for internal storage
architecture rtl_sig of seq_mult_2_fsmd is
type state_type is (rest, iterate, done);
signal state : state_type;
signal rega : unsigned (2*n-1 downto 0);
signal regb : unsigned (n-1 downto 0);
signal regp : unsigned (2*n-1 downto 0);
constant n_zeros : unsigned (n-1 downto 0) := (others => '0');
begin

	-- The next state logic and register in one process
	-- together with the data processing
	process (clk, areset) is
	begin
		-- Asynchronous reset, set reset state, clear all registers
		if areset = '1' then
			state <= rest;
			rega <= (others => '0');
			regb <= (others => '0');
			regp <= (others => '0');
      -- On positive edge, hence flipflops
		elsif rising_edge(clk) then
			case state is
				-- Wait for start to be asserted, then goto iterate
				-- Meanwhile, clear all registers
				when rest => if start = '1' then state <= iterate; end if;
								 rega <= n_zeros & dataa;
								 regb <= datab;
								 regp <= (others => '0');
            -- As long as B is not 0, stay iterating, else we're done
            -- Increment P if lsb of B is 1, left shift A, right shift B
				when iterate => if regb = n_zeros then state <= done; end if;
								    if regb(0) = '1' then regp <= regp + rega; end if;
									 rega <= rega(2*n-2 downto 0) & '0';
				                regb <= '0' & regb(n-1 downto 1);
				-- Wait for start to be de-asserted, then goto rest
				when done => if start = '0' then state <= rest; end if;

				-- If something goes horrible wrong, do nothing...
				when others => null;
			end case;
		end if;
	end process;

	-- Signal ready (combinatorial process)
	ready <= '1' when state = done else '0';

	-- Output the result
	resultp <= regp;

end architecture rtl_sig;
